import asyncio
import zmq.asyncio
import threading
from abc import ABC, abstractmethod

from .conf_logger import logger

class ZmqServerAsync(ABC):
    def __init__(self, host="127.0.0.1", port=5555,
            socket_poll_timeout:int|None=None,
            break_time:float=0.05,
            use_json=True):
        self.host = host
        self.port = port
        self.socket_poll_timeout = socket_poll_timeout
        self.break_time = break_time
        self.use_json = use_json
        self.context = None
        self.socket = None
        self._thread = None
        self._loop = None
        self._stop_event_async = None

    def start(self):
        if self._thread and self._thread.is_alive():
            logger.debug(f"[{self.__class__.__name__}] Already running.")
            return
        self._thread = threading.Thread(target=self._start_event_loop, daemon=True)
        self._thread.start()
        logger.debug(f"[{self.__class__.__name__}] Thread started.")

    def _start_event_loop(self):
        asyncio.set_event_loop(asyncio.new_event_loop())
        self._stop_event_async = asyncio.Event()
        self._loop = asyncio.get_event_loop()
        try:
            self._loop.run_until_complete(self._run())
        finally:
            self._loop.close()
        logger.debug(f"[{self.__class__.__name__}] Event loop closed.")

    async def _run(self):
        self.context = zmq.asyncio.Context()
        self.setup_socket()
        try:
            while not self._stop_event_async.is_set():
                await self._process()
        except Exception as exc:
            logger.warning(f"[{self.__class__.__name__}] Server error: {str(exc)}")
        finally:
            self.socket.close()
            self.context.term()
            logger.debug(f"[{self.__class__.__name__}] Socket closed.")

    async def _process(self):
        if await self.socket_poll(timeout=self.socket_poll_timeout):
            await self.server_loop()
        else:
            if self.break_time > 0:
                await asyncio.sleep(self.break_time)

    def stop(self):
        if self._thread and self._thread.is_alive():
            logger.debug(f"[{self.__class__.__name__}] Stopping server...")
            if self._loop and self._stop_event_async:
                self._loop.call_soon_threadsafe(self._stop_event_async.set)
            self._thread.join(timeout=5)
            logger.debug(f"[{self.__class__.__name__}] Server stopped.")
        else:
            logger.debug(f"[{self.__class__.__name__}] Server has already stopped.")

    def restart(self):
        logger.debug(f"[{self.__class__.__name__}] Restarting server...")
        self.stop()
        self.start()

    @abstractmethod
    def setup_socket(self):
        pass

    @abstractmethod
    async def server_loop(self):
        pass

    @abstractmethod
    async def socket_poll(self, timeout):
        pass



import zmq
import time
import random
import threading

from ..conf_logger import logger

class ZmqLBWorker:
    def __init__(self, endpoint="tcp://localhost:6767",
                 identity=None,
                 message_handler=None):
        self.context = zmq.Context()
        self.worker = self.context.socket(zmq.DEALER)
        self.identity = identity or f"Worker-1"
        self.worker.identity = self.identity.encode()
        self.worker.connect(endpoint)

        self._message_handler = message_handler or (lambda msg, client: print(f"[{self.identity}] Received {msg} from {client}"))

        self._stop_event = threading.Event()
        self._thread = None

    def _run(self):
        logger.debug(f"[{self.__class__.__name__}] is running ...")
        while not self._stop_event.is_set():
            try:
                msg = self.worker.recv_multipart(flags=zmq.NOBLOCK)
                logger.debug(f"[{self.__class__.__name__}] Received from broker: {msg}")
                identity, empty, message = msg
                msg_decoded = message.decode()
                client_id = identity.decode()

                logger.debug(f"[{self.__class__.__name__}] Receive a request '{msg_decoded}' from '{client_id}'")
                time.sleep(random.uniform(0.3, 1.0))  # giả lập xử lý

                logger.debug(f"[{self.__class__.__name__}] Processing")
                self._message_handler(msg_decoded, client_id)

                result = f"{msg_decoded} OK"
                reply = result.encode()
                logger.debug(f"[{self.__class__.__name__}] Reply result '{result}' to '{client_id}'")
                self.worker.send_multipart([identity, b'', reply])
            except zmq.Again:
                time.sleep(0.1)
            except Exception as exc:
                logger.warning(f"[{self.__class__.__name__}] error: {exc}")

    def __call__(self):
        self._stop_event.clear()
        self._thread = threading.Thread(target=self._run)
        self._thread.start()

    def start(self):
        self.__call__()

    def terminate(self):
        self._stop_event.set()
        if self._thread:
            self._thread.join(timeout=2)
        self.worker.close()
        self.context.term()



import logging

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s [%(levelname)s] %(message)s",
    handlers=[ logging.StreamHandler() ]
)

logger = logging.getLogger("apibean.notebook.zmq")
logger.setLevel(logging.DEBUG)



import threading
from collections import deque
from contextlib import redirect_stdout, redirect_stderr

from ._DisplayWriter import DisplayWriter
from ._StreamToDeque import StreamToDeque

class StreamCatcher:
    def __init__(self, max_lines=100, refresh_change=True,
                stdout_prefix='',
                stderr_prefix='',
                redirect_stderr_to_stdout=True):
        self.max_lines = max_lines
        self.redirect_stderr_to_stdout = redirect_stderr_to_stdout

        self._writer = DisplayWriter()

        self._stdout_lock = threading.Lock()
        self._stderr_lock = threading.Lock()

        self.stdout_buffer = deque(maxlen=max_lines)
        self.stderr_buffer = deque(maxlen=max_lines) if not redirect_stderr_to_stdout else \
                self.stdout_buffer

        self._stdout_stream = StreamToDeque(self.stdout_buffer, self._stdout_lock, self._writer,
                max_lines=max_lines,
                refresh_change=refresh_change,
                prefix=stdout_prefix,
                stream_type='stdout')
        self._stderr_stream = StreamToDeque(self.stderr_buffer, self._stderr_lock, self._writer,
                max_lines=max_lines,
                refresh_change=refresh_change,
                prefix=stderr_prefix,
                stream_type='stderr') if not redirect_stderr_to_stdout else self._stdout_stream

        self._stdout_redirect = redirect_stdout(self._stdout_stream)
        self._stderr_redirect = redirect_stderr(self._stderr_stream)

    def __enter__(self):
        self._stdout_redirect.__enter__()
        self._stderr_redirect.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stderr_redirect.__exit__(exc_type, exc_val, exc_tb)
        self._stdout_redirect.__exit__(exc_type, exc_val, exc_tb)

    def get_output(self):
        return list(self.stdout_buffer), list(self.stderr_buffer)


from apibean.notebook._example import assign_example

assign_example(StreamCatcher, '''
import threading
import time

from apibean.notebook.display import StreamCatcher

def example_loop():
    with StreamCatcher(max_lines=8):
        for i in range(20):
            print(f"Printed line {i}")
            time.sleep(0.5)

# Chạy trong thread để không chặn notebook
threading.Thread(target=example_loop, daemon=True).start()
''')



from typing import Literal

from fastapi import APIRouter

HANDLER_TYPE = Literal['ac', 'rr', 'tc']

from .AsyncClient import create_handler_from_router as create_handler_from_router_ac
from .RawResponder import create_handler_from_router as create_handler_from_router_rr
from .TestClient import create_handler_from_router as create_handler_from_router_tc

def create_handler_from_router(router: APIRouter, default_router: APIRouter|None = None,
    handler_type: str | None = 'ac',
):
    if handler_type == 'rr':
        return create_handler_from_router_rr(router, default_router=default_router)

    if handler_type == 'tc':
        return create_handler_from_router_tc(router, default_router=default_router)

    return create_handler_from_router_ac(router, default_router=default_router)



from IPython import get_ipython

example_sources = dict()

def example(module, label: str = "default"):
    # Tra cứu ví dụ trong từ điển
    code = example_sources.get(module, {}).get(label, {}).get("code")
    if code is None:
        raise ValueError(f"Chưa có ví dụ minh họa cho module")

    # Chèn code vào cell tiếp theo
    get_ipython().set_next_input(str(code).strip(), replace=False)


def assign_example(module, code, label: str = "default"):
    example_sources.update({
        module: {
            label: {
                "code": code
            }
        }
    })



import zmq
import json

from typing import Callable, Optional

from .base_server import ZmqServer
from .conf_logger import logger

class ZmqPubSubServer(ZmqServer):
    def __init__(self, *args, message_builder: Optional[Callable[...,dict|str]]=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._message_builder = message_builder
        self._count = 0

    def setup_socket(self):
        self.socket = self.context.socket(zmq.PUB)
        self.socket.bind(f"tcp://{self.host}:{self.port}")
        logger.debug(f"[{self.__class__.__name__}] Bound to {self.host}:{self.port}")

    def server_loop(self):
        result = self.generate_message(self._count)
        topic = result.get("topic", "topic1")
        data = result.get("data", json.dumps({}))
        self.socket.send_multipart([topic.encode(), data.encode()])
        logger.debug(f"[{self.__class__.__name__}] Sent: {data} as '{topic}'")
        self._count += 1

    def socket_poll(self, timeout:int|None):
        return self.socket.poll(timeout=(100 if timeout is None else timeout), flags=zmq.POLLOUT)

    def generate_message(self, count):
        if callable(self._message_builder):
            return self._message_builder(count=count)
        if self.use_json:
            return dict(topic="topic1", data=json.dumps({"task_id": count, "task": f"Task #{count}"}))
        else:
            return dict(topic="topic1", data=f"task_id: {count}")



import asyncio
import json
import zmq

from typing import Awaitable, Callable, Optional

from .base_server_async import ZmqServerAsync
from .conf_logger import logger

class ZmqPushPullServerAsync(ZmqServerAsync):
    def __init__(self, *args, message_builder: Optional[Callable[...,Awaitable[dict|str]]]=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._message_builder = message_builder
        self._count = 0

    def setup_socket(self):
        self.socket = self.context.socket(zmq.PUSH)
        self.socket.bind(f"tcp://{self.host}:{self.port}")
        logger.debug(f"[{self.__class__.__name__}] Bound to tcp://{self.host}:{self.port}")

    async def server_loop(self):
        result = await self.generate_message(self._count)
        data = result.get("data", json.dumps({}))
        await self.socket.send_string(data)
        logger.debug(f"[{self.__class__.__name__}] Sent: {data}")
        self._count += 1

    async def socket_poll(self, timeout:int|None):
        return await self.socket.poll(timeout=(100 if timeout is None else timeout), flags=zmq.POLLOUT)

    async def generate_message(self, count):
        if callable(self._message_builder):
            return await self._message_builder(count=count)
        if self.use_json:
            return dict(data=json.dumps({"task_id": count, "task": f"Task #{count}"}))
        else:
            return dict(data=f"task_id: {count}")



import asyncio
import zmq
import json

from typing import Awaitable, Callable, Optional

from .base_server_async import ZmqServerAsync
from .conf_logger import logger

class ZmqReqRepServerAsync(ZmqServerAsync):
    def __init__(self, *args, message_handler: Optional[Callable[...,Awaitable[dict|str]]]=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._message_handler = message_handler

    def setup_socket(self):
        self.socket = self.context.socket(zmq.REP)
        self.socket.bind(f"tcp://{self.host}:{self.port}")
        logger.debug(f"[{self.__class__.__name__}] Bound to tcp://{self.host}:{self.port}")

    async def server_loop(self):
        msg = await self.socket.recv_string()
        logger.debug(f"[{self.__class__.__name__}] Received: {msg}")
        result = await self.process_message(data=msg)
        data = result.get("data", json.dumps({}))
        await self.socket.send_string(str(data))

    async def socket_poll(self, timeout:int):
        return await self.socket.poll(timeout=(1000 if timeout is None else timeout))

    async def process_message(self, data) -> dict:
        logger.debug(f"[{self.__class__.__name__}] Processing: {data}")
        if self._message_handler:
            return await self._message_handler(data=data)
        if self.use_json:
            return dict(data=json.dumps({"echo": data}))
        else:
            return dict(data=f"Echo: {data}")



import asyncio

from fastapi import APIRouter, FastAPI, Request
from fastapi.responses import Response
from fastapi.testclient import TestClient

def create_handler_from_router(router: APIRouter, default_router: APIRouter|None = None):
    app = FastAPI()
    app.include_router(router)
    if default_router is not None:
        app.include_router(default_router)

    client = TestClient(app)

    async def handler(req: Request) -> Response:
        body = await req.body()
        headers = dict(req.headers)
        url = req.url.path

        # TestClient only sync, so run in thread pool
        def do_request():
            method = req.method.lower()
            fn = getattr(client, method)
            body_args = dict()
            if method in ['post', 'put', 'patch']:
                body_args.update(data=body)
            return fn(url, headers=headers, params=req.query_params, **body_args)

        loop = asyncio.get_event_loop()
        response = await loop.run_in_executor(None, do_request)

        return Response(
            content=response.content,
            status_code=response.status_code,
            headers=dict(response.headers),
        )

    return handler



import zmq
import time
import threading

from ..conf_logger import logger

class ZmqLBClient:
    def __init__(self, endpoint="tcp://localhost:6666",
                 identity=None,
                 message_generator=None,
                 message_processor=None,
                 total_messages=5):
        self.context = zmq.Context()
        self.client = self.context.socket(zmq.DEALER)
        self.identity = identity or f"Client-{time.time()}"
        self.client.identity = self.identity.encode()
        self.client.connect(endpoint)

        self._message_generator = message_generator or (lambda i: f"Job-{i}")
        self._message_processor = message_processor or (lambda reply: print(f"[{self.identity}] Got reply: {reply}"))
        self.total_messages = total_messages

        self._stop_event = threading.Event()
        self._thread = None

    def _run(self):
        logger.debug(f"[{self.__class__.__name__}] is running ...")
        for i in range(self.total_messages):
            if self._stop_event.is_set():
                break
            msg = self._message_generator(i)
            self.client.send_multipart([b'', msg.encode()])
            time.sleep(0.1)

        for _ in range(self.total_messages):
            if self._stop_event.is_set():
                break
            try:
                msg = self.client.recv_multipart(flags=zmq.NOBLOCK)
                logger.debug(f"[{self.__class__.__name__}] Received from broker: {msg}")
                empty, reply = msg
                self._message_processor(reply.decode())
            except zmq.Again:
                time.sleep(0.1)
            except Exception as exc:
                logger.warning(f"[{self.__class__.__name__}] error: {exc}")

    def __call__(self):
        self._thread = threading.Thread(target=self._run)
        self._thread.start()

    def start(self):
        self.__call__()

    def terminate(self):
        self._stop_event.set()
        if self._thread:
            self._thread.join(timeout=2)
        self.client.close()
        self.context.term()



from .fastapi_server import EmbeddedAPIServer
from .fastapi_server import EmbeddedAPIServerLab as EmbeddedAPIService

from .handlers import create_handler_from_router
from .fastapi_util import default_handler
from .fastapi_util import default_router



import threading
import asyncio

from typing import Callable, Optional

from fastapi import APIRouter, FastAPI, Request

import uvicorn

from .handlers import HANDLER_TYPE
from .handlers import create_handler_from_router

from .fastapi_util import HTTP_METHODS
from .fastapi_util import default_handler
from .fastapi_util import default_router

from .middlewares.PrettyJSONMiddleware import PrettyJSONMiddleware

from ..utils.net_util import is_port_available

class EmbeddedAPIServerBase:
    def __init__(self, host="0.0.0.0", port=8000, log_level:str="error",
            control_base_path: str = "__notebook__",
            pretty_json_response: bool = True):
        # service settings
        self.host = host
        self.port = port
        self.log_level = log_level
        self.control_base_path = control_base_path
        self.pretty_json_response = pretty_json_response
        # private properties
        self._thread: Optional[threading.Thread] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._server: Optional[uvicorn.Server] = None
        self._running: bool = False
        self._handler = None

    def _create_app(self) -> FastAPI:
        app = FastAPI()

        if self.pretty_json_response:
            app.add_middleware(PrettyJSONMiddleware)

        @app.get(f"/{self.control_base_path}/status")
        async def status_endpoint():
            return self.status()

        @app.api_route("/{full_path:path}", methods=HTTP_METHODS, include_in_schema=False)
        async def main_handler(request: Request, full_path: str):
            if self._handler:
                return await self._handler(request)
            return await default_handler(request, full_path, handler_name="universal")

        return app

    def _run_server(self):
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        app = self._create_app()
        config = uvicorn.Config(app, host=self.host, port=self.port, log_level=self.log_level,
                loop="asyncio")
        self._server = uvicorn.Server(config)

        self._running = True
        try:
            self._loop.run_until_complete(self._server.serve())
        finally:
            self._loop.close()
        self._running = False

    def _print_info(self, msg):
        print(msg)

    def start(self):
        if self._running:
            self._print_info("⚠️ Server is already running.")
            return

        if not is_port_available(port=self.port, host=self.host):
            self._print_info(f"❌ Address ('{self.host}', {self.port}) already in use")
            return

        self._thread = threading.Thread(target=self._run_server, daemon=True)
        self._thread.start()
        self._print_info(f"✅ Server started at http://{self.host}:{self.port}")

    def stop(self):
        if self._server:
            self._print_info("🛑 Stopping server...")
            self._server.should_exit = True
            if self._thread and self._thread.is_alive():
                self._thread.join(timeout=5)
            self._thread = None
            self._server = None
            self._loop = None
        else:
            self._print_info("⚠️ Server not running.")

    def restart(self):
        self.stop()
        self.start()

    def status(self) -> dict:
        return {
            "running": self._running,
            "thread_alive": self._thread.is_alive() if self._thread else False,
            "host": self.host,
            "port": self.port,
            "handler_set": self._handler is not None
        }

    def update_handler(self, func: Optional[Callable[..., dict]]):
        self._handler = func
        self._print_info("✅ Handler updated.")

    def update_router(self, router: Optional[APIRouter], handler_type: HANDLER_TYPE | None = None):
        handler = create_handler_from_router(router,
                default_router=default_router,
                handler_type=handler_type) if router is not None else None
        self.update_handler(handler)



import zmq
import json
import time

from .conf_logger import logger

class ZmqReqRepClient:
    def __init__(self, host="127.0.0.1", port=5555, retries=5, retry_interval=1.0,
            return_none_on_error: bool = False):
        self.host = host
        self.port = port
        self.retries = retries
        self.retry_interval = retry_interval
        self.return_none_on_error = return_none_on_error

    def __call__(self, msg: str) -> str|None:
        endpoint = f"tcp://{self.host}:{self.port}"

        reply = None
        for attempt in range(self.retries):
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.setsockopt(zmq.LINGER, 0)
            socket.connect(endpoint)

            try:
                logger.debug(f"[{self.__class__.__name__}] Attempt {attempt+1}: Sending '{msg}'")
                socket.send_string(str(msg))
                poller = zmq.Poller()
                poller.register(socket, zmq.POLLIN)
                socks = dict(poller.poll(timeout=2000))

                if socks.get(socket) == zmq.POLLIN:
                    reply = socket.recv_string()
                    logger.debug(f"[{self.__class__.__name__}] Reply: {reply}")
                    break
                else:
                    logger.debug(f"[{self.__class__.__name__}] No response, retrying...")
            except zmq.ZMQError as e:
                logger.debug(f"[{self.__class__.__name__}] ZMQ Error: {e}, retrying...")
            finally:
                socket.close()
                context.term()

            time.sleep(self.retry_interval)

        if reply is None:
            logger.debug(f"[{self.__class__.__name__}] Failed to connect after retries.")
            if self.return_none_on_error:
                return None
            raise Exception(f"Failed to connect after {self.retries} attempt retries")

        return reply


class ZmqReqRepJsonClient(ZmqReqRepClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, data: dict) -> dict|None:
        msg = json.dumps(data)

        reply = super().__call__(msg)

        if reply is None:
            return reply

        if not isinstance(reply, str):
            return reply

        try:
            reply = json.loads(reply)
        except json.JSONDecodeError as jde:
            logger.debug(f"[{self.__class__.__name__}] JSON decode error")
            if self.return_none_on_error:
                return None
            raise jde

        return reply



import threading

from .tail_log_file import tail_log_file

class TailFileService:
    def __init__(self, filepath, max_lines=100, interval=1.0, display_writer=None):
        self.filepath = filepath
        self.max_lines = max_lines
        self.interval = interval
        self._thread = None
        self._display_writer = display_writer
        self._stop_event = threading.Event()

    def _run(self):
        tail_log_file(
            self.filepath,
            max_lines=self.max_lines,
            interval=self.interval,
            display_writer=self._display_writer,
            stop_event=self._stop_event
        )

    def _print(self, msg):
        print(msg)

    def start(self, filepath=None, max_lines=None, interval=None):
        if self._thread and self._thread.is_alive():
            self._print(f"{self.__class__.__name__} is already running.")
            return

        if filepath is not None:
            self.filepath = filepath
        if max_lines is not None:
            self.max_lines = max_lines
        if interval is not None:
            self.interval = interval

        if self.filepath is None:
            self._print(f"{self.__class__.__name__} is not ready (filepath is None)")
            return

        self._stop_event.clear()
        self._thread = threading.Thread(target=self._run, daemon=True)
        self._thread.start()
        self._print(f"{self.__class__.__name__} started.")

    def stop(self):
        if not self._thread:
            self._print(f"{self.__class__.__name__} is not running.")
            return

        self._stop_event.set()
        self._thread.join()
        self._thread = None
        self._print(f"{self.__class__.__name__} stopped.")



import threading
import time
import zmq
from abc import ABC, abstractmethod

from .conf_logger import logger

class ZmqServer(ABC):
    def __init__(self, host="127.0.0.1", port=5555,
            socket_poll_timeout:int|None=None,
            break_time:float=0.05,
            use_json=True):
        self.host = host
        self.port = port
        self.socket_poll_timeout = socket_poll_timeout
        self.break_time = break_time
        self.use_json = use_json
        self.context = None
        self.socket = None
        self._stop_event = threading.Event()
        self._thread = None

    def start(self):
        if self._thread and self._thread.is_alive():
            logger.debug(f"[{self.__class__.__name__}] Server already running.")
            return
        logger.debug(f"[{self.__class__.__name__}] Starting server thread...")
        self._stop_event.clear()
        self.context = zmq.Context()
        self.setup_socket()
        self._thread = threading.Thread(target=self._target, daemon=True)
        self._thread.start()
        logger.debug(f"[{self.__class__.__name__}] Server started.")

    def stop(self):
        logger.debug(f"[{self.__class__.__name__}] Stopping server thread...")
        self._stop_event.set()
        if self._thread:
            self._thread.join(timeout=5)
        if self.socket:
            self.socket.close()
        if self.context:
            self.context.term()
        logger.debug(f"[{self.__class__.__name__}] Server stopped.")

    def restart(self):
        logger.debug(f"[{self.__class__.__name__}] Restarting server...")
        self.stop()
        self.start()

    def _target(self):
        while not self._stop_event.is_set():
            try:
                self._process()
            except zmq.ZMQError as e:
                if e.errno == zmq.ETERM:
                    logger.debug(f"[{self.__class__.__name__}] Context terminated.")
                elif e.errno == zmq.ENOTSOCK:
                    logger.debug(f"[{self.__class__.__name__}] Invalid socket.")
                else:
                    raise

    def _process(self):
        if self.socket_poll(timeout=self.socket_poll_timeout):
            self.server_loop()
        elif self.break_time > 0:
            time.sleep(self.break_time)

    @abstractmethod
    def setup_socket(self):
        pass

    @abstractmethod
    def server_loop(self):
        pass

    @abstractmethod
    def socket_poll(self, timeout):
        pass



from IPython.display import clear_output, display, update_display, HTML

class DisplayWriter:
    surround_top: str = "⬆️<br>"
    surround_bottom: str = "<br>⬇️"

    def __init__(self, title:str="stdout stream", use_display_id=True):
        if use_display_id:
            self._display_id = display(self._get_formatted_title(title), display_id=True)
        else:
            self._display_id = None

    def render(self, lines, stream_type:str|None=None):
        html_object = self._get_formatted_lines(lines, stream_type=stream_type)
        if self._display_id:
            update_display(html_object, display_id=self._display_id.display_id)
        else:
            clear_output(wait=True)
            display(html_object)

    def write(self, html:str):
        html_object = self._get_enclosed_html(html)
        if self._display_id:
            update_display(html_object, display_id=self._display_id.display_id)
        else:
            display(html_object)

    def clear(self):
        if self._display_id:
            blank = HTML("<pre style='margin:0'></pre>")
            update_display(blank, display_id=self._display_id.display_id, clear=True)
        else:
            clear_output(wait=True)

    def _get_formatted_title(self, title):
        return HTML(f"<pre style='margin:0;color:#555'>[{title}]</pre>")

    def _get_formatted_lines(self, lines, stream_type:str|None=None):
        if stream_type == 'stderr':
            html_lines = []
            for line in lines:
                html_lines.append(f"<span style='color:red'>{line}</span>")
            html = "<br>".join(html_lines)
        else:
            html = '<br>'.join(lines)
        return self._get_enclosed_html(html)

    def _get_enclosed_html(self, html):
        return HTML(f"<pre style='margin:0'>{self.surround_top}{html}{self.surround_bottom}</pre>")



import httpx
from fastapi import APIRouter, FastAPI, Request, Response

METHODS_WITH_BODY = {"POST", "PUT", "PATCH", "DELETE"}    

def create_handler_from_router(router: APIRouter, default_router: APIRouter|None = None):
    app = FastAPI()
    app.include_router(router)
    if default_router is not None:
        app.include_router(default_router)

    client = httpx.AsyncClient(transport=httpx.ASGITransport(app=app),
        base_url="http://internal")  # httpx >= 0.23

    async def handler(request: Request) -> Response:
        method = request.method.upper()
        opts = dict(
            method = method,
            url = request.url.path,
            headers = dict(request.headers),
            params = dict(request.query_params),
        )
        if method in METHODS_WITH_BODY:
            body = await request.body()
            opts.update(content=body)
    
        resp = await client.request(**opts)
    
        return Response(
            content=resp.content,
            status_code=resp.status_code,
            headers=dict(resp.headers)
        )

    return handler



import socket

def is_port_available(port: int, host: str="0.0.0.0") -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(1)
        try:
            s.bind((host, port))
            return True
        except OSError:
            return False


def get_process_using_port(port: int):
    import psutil
    for conn in psutil.net_connections(kind="inet"):
        if conn.laddr.port == port:
            return psutil.Process(conn.pid).name() if conn.pid else None
    return None



import threading
import zmq

from typing import Callable, Optional

from .conf_logger import logger

class ZmqPushPullClient:
    def __init__(self, message_handler: Optional[Callable]=None,
            host="127.0.0.1", port=5555, timeout=5000):
        self.host = host
        self.port = port
        self.timeout = timeout
        self._message_handler = message_handler
        self._stop_event = threading.Event()
        self._thread = None

    def __call__(self):
        if self._thread and self._thread.is_alive():
            logger.debug(f"[{self.__class__.__name__}] Already running.")
            return
        self._stop_event.clear()
        self._thread = threading.Thread(target=self._run_loop, daemon=True)
        self._thread.start()

    def _run_loop(self):
        ctx = zmq.Context()
        sock = ctx.socket(zmq.PULL)
        sock.connect(f"tcp://{self.host}:{self.port}")
        logger.debug(f"[{self.__class__.__name__}] Waiting for messages...")
        poller = zmq.Poller()
        poller.register(sock, zmq.POLLIN)

        try:
            while not self._stop_event.is_set():
                socks = dict(poller.poll(self.timeout))
                if socks.get(sock) == zmq.POLLIN:
                    data = sock.recv_string()
                    logger.debug(f"[{self.__class__.__name__}] Got: {data}")
                    if callable(self._message_handler):
                        self._message_handler(data) # (data=data)
                else:
                    logger.debug(f"[{self.__class__.__name__}] No message yet...")
        except Exception as exc:
            logger.debug(f"[{self.__class__.__name__}] Error: {exc}")
        finally:
            sock.close()
            ctx.term()
            logger.debug(f"[{self.__class__.__name__}] Cleaned up.")

    def terminate(self):
        logger.debug(f"[{self.__class__.__name__}] Terminating...")
        self._stop_event.set()
        if self._thread:
            self._thread.join()
            logger.debug(f"[{self.__class__.__name__}] Thread joined.")



import os
import socket
import psutil

from IPython import get_ipython

from ._EmbeddedAPIServerDev import EmbeddedAPIServerDev

class EmbeddedAPIServerLab(EmbeddedAPIServerDev):
    def __init__(self, *args, show_kernel_info: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.show_kernel_info = show_kernel_info

    def status(self) -> dict:
        info = super().status()

        if not self.show_kernel_info:
            return info

        # IPython kernel info
        ipy = get_ipython()
        shell_class = ipy.__class__.__name__ if ipy else None
        shell_module = ipy.__class__.__module__ if ipy else None
        process = psutil.Process(os.getpid())

        # Network
        hostname = socket.gethostname()
        try:
            local_ip = socket.gethostbyname(hostname)
        except:
            local_ip = "Unavailable"

        # Try get Jupyter info from env
        jupyter_env = {
            k: v for k, v in os.environ.items()
            if k.startswith("JPY_") or k.startswith("JUPYTER") or k in ("VIRTUAL_ENV", "CONDA_DEFAULT_ENV")
        }

        return {
            **info,
            "kernel": {
                "pid": os.getpid(),
                "parent_pid": process.ppid(),
                "shell_class": shell_class,
                "shell_module": shell_module,
            },
            "jupyter_env": jupyter_env,
        }



from fastapi import Request, Response
from starlette.middleware.base import BaseHTTPMiddleware
import json

import logging
_logger = logging.getLogger(__name__)

class PrettyJSONMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        response = await call_next(request)

        # Chỉ xử lý nếu có header yêu cầu pretty JSON
        if request.headers.get("X-Pretty-Json-Response"):
            _logger.debug("Header[X-Pretty-Json-Response] is effected")
            _logger.debug(f"Type of response: {type(response)}")

            # Chỉ xử lý nếu Content-Type là JSON
            content_type = response.headers.get("Content-Type", "")
            if "application/json" in content_type:

                # Đọc body hiện tại từ response
                body = b""
                async for chunk in response.body_iterator:
                    body += chunk
                _logger.debug(f"Response body has read")

                try:
                    # Parse JSON → format lại → encode lại
                    raw_data = json.loads(body)
                    pretty_body = json.dumps(raw_data, indent=4, ensure_ascii=False)
                    new_body = pretty_body.encode("utf-8")
                    _logger.debug(f"Parsing body is ok")
                except Exception as exc:
                    # Nếu không phải JSON hoặc lỗi parse → giữ nguyên
                    _logger.debug(f"Parsing body error: {exc}")
                    return response

                # Trả lại response mới với body đã format
                _logger.debug(f"Return new formatted Response")
                return Response(
                    content=new_body,
                    status_code=response.status_code,
                    headers={k: v for k, v in response.headers.items() if k.lower() != "content-length"},
                    media_type="application/json"
                )

        return response



from typing import Callable, Awaitable

from fastapi import FastAPI, Request
from fastapi.routing import APIRouter
from starlette.responses import Response
from starlette.types import Receive, Scope, Send


def create_handler_from_router(router: APIRouter, default_router: APIRouter|None = None) \
        -> Callable[[Request], Awaitable[Response]]:
    # Gắn router vào một FastAPI app tạm thời
    temp_app = FastAPI()
    temp_app.include_router(router)
    if default_router is not None:
        temp_app.include_router(default_router)

    async def handler(req: Request) -> Response:
        # Tạo response container
        send_buffer = {}

        async def send(message):
            send_buffer["message"] = message

        # Run app
        responder = temp_app  # FastAPI app là một ASGI app
        await responder(req.scope, req.receive, send)

        # Trích response từ message
        message = send_buffer.get("message")
        if message is None:
            return Response("No response", status_code=500)

        # Nếu là response start, thì ok
        if message["type"] == "http.response.start":
            # Nội dung thực sẽ được gửi qua "http.response.body"
            body_msg = send_buffer.get("body_message")
            body = body_msg["body"] if body_msg else b""
            return Response(content=body, status_code=message["status"])
        elif message["type"] == "http.response.body":
            return Response(content=message["body"], status_code=200)

        return Response("Unhandled response", status_code=500)

    return handler



import threading
import time
from collections import deque
from IPython.utils.capture import capture_output

from ..display import StreamToDeque, DisplayWriter
from ._IntervalRunner_StreamCatcher import IntervalRunner as OriginIntervalRunner

class IntervalRunner(OriginIntervalRunner):
    def __init__(self, *args, max_lines=10, **kwargs):
        super().__init__(*args, max_lines=max_lines, **kwargs)
        self.stdout_buffer = deque(maxlen=max_lines)
        self._stdout_lock = threading.Lock()
        self.writer = DisplayWriter(title="IntervalRunner's output")
        self.stream = StreamToDeque(self.stdout_buffer, self._stdout_lock, self.writer,
                max_lines=max_lines,
                refresh_change=True,
                prefix='',
                stream_type=None)

    def _loop(self):
        while self._running:
            with capture_output() as captured:
                self.func()
            lines = captured.stdout.strip().splitlines()
            for line in lines:
                self.stream.write(line)
            '''
            IOPub message rate exceeded.
            The Jupyter server will temporarily stop sending output
            to the client in order to avoid crashing it.
            To change this limit, set the config variable
            `--ServerApp.iopub_msg_rate_limit`.

            Current values:
            ServerApp.iopub_msg_rate_limit=1000.0 (msgs/sec)
            ServerApp.rate_limit_window=3.0 (secs)
            '''
            if self.interval > 0:
                time.sleep(self.interval)

    def clear(self):
        if self.writer:
            self.writer.clear()


from apibean.notebook._example import assign_example

assign_example(IntervalRunner, '''
import random

from apibean.notebook.runners import IntervalRunner

counter = 0

def my_func():
    global counter
    counter += 1
    print(f"[{counter}] Random number: {random.randint(1, 100)}")

runner = IntervalRunner(my_func, interval=0.7, max_lines=10)

# <- Press [ Ctrl + Shift + - ] to split cells
runner.start()

# <- Press [ Ctrl + Shift + - ] to split cells
runner.stop()
''')



import os
import platform
import psutil

from ._EmbeddedAPIServerBase import EmbeddedAPIServerBase

class EmbeddedAPIServerDev(EmbeddedAPIServerBase):
    def __init__(self, *args, show_system_info: bool = True, **kwargs):
        self.show_system_info = show_system_info
        super().__init__(*args, **kwargs)

    def status(self) -> dict:
        info = super().status()

        if not self.show_system_info:
            return info

        cpu_percent = psutil.cpu_percent(interval=0.1)
        virtual_mem = psutil.virtual_memory()
        process = psutil.Process(os.getpid())
        memory_info = process.memory_info()

        return {
            **info,
            "system": {
                "platform": platform.system(),
                "cpu_percent": cpu_percent,
                "memory": {
                    "total": virtual_mem.total,
                    "available": virtual_mem.available,
                    "used": virtual_mem.used,
                    "used_by_process": memory_info.rss,
                    "percent": virtual_mem.percent,
                }
            }
        }



from fastapi import APIRouter, Request

HTTP_METHODS = ["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "HEAD"]

default_router = APIRouter()

@default_router.api_route("/{full_path:path}", methods=HTTP_METHODS, include_in_schema=False)
async def default_handler(request: Request, full_path: str, handler_name: str = "default"):
    """
    Default catch-all request handler used as a fallback in the API proxy/router.

    This handler is registered with a wildcard path (``/{full_path:path}``) and
    supports all common HTTP methods. It is intended to act as a *last-resort*
    handler when no specific route or extension matches the incoming request.

    Typical use cases:
    - Development-time API proxying / MITM inspection
    - Debugging and introspecting incoming HTTP requests
    - Serving as a placeholder handler before implementing real endpoints
    - Returning a normalized view of the request for notebooks or tooling

    Behavior:
    - Attempts to parse the request body as JSON; if parsing fails, ``body`` is set to ``None``
    - Collects and returns key request attributes in a structured dictionary:
      - handler name
      - resolved request path
      - HTTP method
      - query parameters
      - headers
      - request body (if JSON)

    Notes:
    - This handler does NOT forward the request upstream.
    - It should typically be placed at the end of the routing chain.
    - ``include_in_schema=False`` ensures it does not appear in OpenAPI docs.

    Parameters
    ----------
    request : fastapi.Request
        The incoming HTTP request object.
    full_path : str
        The full unmatched request path captured by the wildcard route.
    handler_name : str, optional
        Logical name of the handler, useful for debugging or tracing
        (default: "default").

    Returns
    -------
    dict
        A dictionary containing a structured representation of the incoming request.
    """
    try:
        body = await request.json()
    except:
        body = None
    return dict(
        handler_name=handler_name,
        path=full_path,
        method=request.method,
        query=dict(request.query_params),
        headers=dict(request.headers),
        body=body
    )



from collections import deque

class StreamToDeque:
    def __init__(self, buffer: deque, lock, writer, max_lines=10, refresh_change=True, prefix='', stream_type='stdout'):
        self.buffer = buffer
        self.lock = lock
        self.writer = writer
        self.max_lines = max_lines
        self.refresh_change = refresh_change
        self.prefix = prefix
        self.stream_type = stream_type

    def write(self, message):
        with self.lock:
            lines = message.splitlines()
            for line in lines:
                if line.strip():
                    self.buffer.append(f"{self.prefix}{line}")
            if self.refresh_change:
                self._refresh_output()

    def flush(self):
        pass

    def _refresh_output(self):
        self.writer.render(self.buffer, stream_type=self.stream_type)



import zmq
import threading

from ..conf_logger import logger

class ZmqLBBroker:
    def __init__(self, frontend_port=6666, backend_port=6767):
        self.context = zmq.Context()

        self.frontend = self.context.socket(zmq.ROUTER)
        self.backend = self.context.socket(zmq.ROUTER)

        self.backend.setsockopt(zmq.ROUTER_MANDATORY, 1)

        self.frontend.bind(f"tcp://*:{frontend_port}")
        logger.debug(f"[{self.__class__.__name__}] Bound frontend to *:{frontend_port}")

        self.backend.bind(f"tcp://*:{backend_port}")
        logger.debug(f"[{self.__class__.__name__}] Bound backend to *:{backend_port}")

        self._poller = zmq.Poller()
        self._poller.register(self.frontend, zmq.POLLIN)
        self._poller.register(self.backend, zmq.POLLIN)

        self._stop_event = threading.Event()
        self._thread = None

    def _run(self):
        logger.debug(f"[{self.__class__.__name__}] is running ...")
        while not self._stop_event.is_set():
            try:
                socks = dict(self._poller.poll(timeout=100))
                if self.frontend in socks:
                    logger.debug(f"[{self.__class__.__name__}] Receive a request from frontend")
                    msg = self.frontend.recv_multipart()
                    logger.debug(f"[{self.__class__.__name__}] Received from frontend: {msg}")
                    logger.debug(f"[{self.__class__.__name__}] Dispatch the request to backend")
                    self.backend.send_multipart([b'Worker-1'] + msg)

                if self.backend in socks:
                    logger.debug(f"[{self.__class__.__name__}] Receive a reply from the backend")
                    msg = self.backend.recv_multipart()
                    logger.debug(f"[{self.__class__.__name__}] Received from backend: {msg}")
                    worker_id, *payload = msg
                    logger.debug(f"[{self.__class__.__name__}] Dispatch the reply to frontend")
                    self.frontend.send_multipart(payload)
            except zmq.ZMQError as exc:
                logger.warning(f"[{self.__class__.__name__}] Broker error: {str(exc)}")
                break

    def __call__(self):
        self._thread = threading.Thread(target=self._run)
        self._thread.start()

    def start(self):
        self.__call__()

    def terminate(self):
        self._stop_event.set()
        if self._thread:
            self._thread.join(timeout=2)
        self.frontend.close()
        self.backend.close()
        self.context.term()



from . import lb

from .base_server import ZmqServer

from .pub_sub_client import ZmqPubSubClient
from .pub_sub_server_async import ZmqPubSubServerAsync
from .pub_sub_server import ZmqPubSubServer

from .push_pull_client import ZmqPushPullClient
from .push_pull_server_async import ZmqPushPullServerAsync
from .push_pull_server import ZmqPushPullServer

from .req_rep_client import ZmqReqRepClient
from .req_rep_server_async import ZmqReqRepServerAsync
from .req_rep_server import ZmqReqRepServer

__all__ = [
    "lb",
    "ZmqServer",
    "ZmqPubSubServer",
    "ZmqPubSubServerAsync",
    "ZmqPubSubClient",
    "ZmqPushPullServer",
    "ZmqPushPullServerAsync",
    "ZmqPushPullClient",
    "ZmqReqRepServer",
    "ZmqReqRepServerAsync",
    "ZmqReqRepClient",
]



import os
import time
from collections import deque
from IPython.display import clear_output

from ..display import DisplayWriter

def _read_last_lines_by_seek(filepath, max_lines=100, encoding='utf-8'):
    """
    Đọc file từ cuối, tìm và trả về max_lines dòng cuối cùng.
    """
    with open(filepath, 'rb') as f:
        f.seek(0, os.SEEK_END)
        pos = f.tell()
        block_size = 1024
        data = bytearray()
        line_count = 0

        while pos > 0 and line_count <= max_lines:
            read_size = min(block_size, pos)
            pos -= read_size
            f.seek(pos)
            block = f.read(read_size)
            data = block + data
            line_count = data.count(b'\n')

        lines = data.split(b'\n')[-max_lines:]
        return [line.decode(encoding, errors='replace') for line in lines]

def _print_deque_buffer(log_buffer, display_writer=None, print_with_join:bool=False):
    if display_writer:
        _print_deque_buffer_to_display(log_buffer, display_writer=display_writer)
        return
    _print_deque_buffer_to_console(log_buffer, print_with_join=print_with_join)

def _print_deque_buffer_to_display(log_buffer, display_writer):
    display_writer.render(log_buffer)

def _print_deque_buffer_to_console(log_buffer, print_with_join:bool=False):
    clear_output(wait=True)
    if print_with_join:
        print("\n".join(log_buffer))
        return
    for line in log_buffer:
        print(line)

def tail_log_file(filepath, max_lines=100, interval=1.0, follow=True,
        encoding='utf-8',
        display_writer=None,
        create_display_writer_if_not_found=True,
        stop_event=None):
    """
    Đọc log như tail -f, có hỗ trợ dừng bằng stop_event (nếu cung cấp).
    """
    try:
        log_buffer = deque(
            _read_last_lines_by_seek(filepath, max_lines, encoding),
            maxlen=max_lines
        )

        if display_writer is None and create_display_writer_if_not_found:
            display_writer = DisplayWriter(title=f"{tail_log_file.__name__}'s output")
        _print_deque_buffer(log_buffer, display_writer=display_writer)

        if not follow:
            return

        with open(filepath, 'r', encoding=encoding, errors='replace') as f:
            f.seek(0, os.SEEK_END)

            while not (stop_event and stop_event.is_set()):
                line = f.readline()
                if line:
                    log_buffer.append(line.rstrip())
                    _print_deque_buffer(log_buffer, display_writer=display_writer)
                else:
                    time.sleep(interval)

        print(f"{tail_log_file.__name__}('{filepath}') ended.")
    except KeyboardInterrupt:
        print(f"{tail_log_file.__name__}('{filepath}') has stopped.")
    except Exception as e:
        print(f"{tail_log_file.__name__} error: {e}")



import threading
import time

from ..display import StreamCatcher

class IntervalRunner:
    def __init__(self, func, interval=0.1, max_lines=10):
        self.func = func
        self.interval = interval
        self.max_lines = max_lines
        self._running = False
        self._thread = None

    def _loop(self):
        with StreamCatcher(max_lines=self.max_lines):
            while self._running:
                self.func()
                time.sleep(self.interval)

    def start(self):
        if not self._running:
            self._running = True
            self._thread = threading.Thread(target=self._loop)
            self._thread.daemon = True
            self._thread.start()

    def stop(self):
        self._running = False
        if self._thread:
            self._thread.join()
        self._thread = None
